package Plugins::MyVOCO::Utils;

use JSON::XS::VersionOneAndTwo;
use LWP::Simple;
use URI;
use Data::Dumper;

use Slim::Utils::Cache;
use Slim::Utils::Log;
use Slim::Utils::Prefs;
use Slim::Networking::SimpleAsyncHTTP;

my $log   = logger('plugin.myvoco');
my $prefs = preferences('plugin.myvoco');
my $cache = Slim::Utils::Cache->new('voco');
my $httptimeout = $prefs->get('timeout', 3);

sub buildUrlForVzc {
	my ($clientIP, $method, $vocoUrl)  = @_;
	$log->debug('ip' . $clientIP);
	$log->debug('method' . $method);
	$log->debug('url' . $vocoUrl);
	my $pUrl = URI->new($vocoUrl, 'voco');
	my $service = substr($pUrl->path(), 1); # substr($, 1) removes the preceding '/' from path
	$log->debug('service [' . $service . ']');
	my $methodString = $method ? 'm=' . $method . '&' : '' ;
	$log->debug('method [' . $methodString . ']');

	my $vzcURL = 'http://' . $clientIP . '/vzc/cgi?command=service&' . 
			$methodString .
			's=' . $service . '&' . $pUrl->query();
	$log->debug('vzcURL  [' . $vzcURL . ']' );

	return $vzcURL;
}

sub getVzcResponse {
	# args are: $clientIP, $method, $vocoUrl
	my $vzcURL = buildUrlForVzc(@_);
	my $content = get($vzcURL);
	return $content;
}

sub callVzcAsync {
	my ($clientIP, $method, $vocoUrl, $cb, $ecb, $params)  = @_;
	$log->debug('ip' . $clientIP);
	$log->debug('method' . $method);
	$log->debug('url' . $vocoUrl);
	$log->debug('cb' . $cb);
	$log->debug('ecb' . $ecb);
	$log->debug('params' . $params);
	my $vzcURL = buildUrlForVzc($clientIP, $method, $vocoUrl);
	$log->info('async call for [' . $vzcURL . ']');
	eval {
		my $http = Slim::Networking::SimpleAsyncHTTP->new(
			$cb,
			$ecb,
			$params,
		);
		$http->get($vzcURL);
	};
	if ($@) {
		$log->error('something bad happened');
		$log->error($@);
	}
}

my %cacheInfo = (	'metadata' 			=> {	'prefix' => 'voco_url_meta',
												'duration' => 86400 },
					'player_control' 	=> {	'prefix' => 'voco_url_control',
												'duration'=> 3600 }
	);

my @caching = ();

sub getFromCacheOrRemote {
	my $type = shift;
	my $player = shift;
	my $url = shift;
	my $block = shift;
	
	$log->debug('prefix [' . $cacheInfo{$type}{'prefix'} . '] for type [' . $type . ']');
	my $master = $player->master();
	my $hashkey = $cacheInfo{$type}{prefix} . $master->id . $url;
	$log->debug('hashkey [' . $hashkey . ']');
	my $cachedData = $cache->get($hashkey);
	if (defined $cachedData) {
		$log->debug('hashkey was found');
		return $cachedData
	} else {
		$log->debug("nothing cached with key $hashkey block [$block]");
		my $ip = $master->ip();
		if ($block) {
			### We should block, let's get the response, process everything and respond.
			my $content = getVzcResponse($ip, $type, $url);
			my $resp = 0;
			eval { $resp = decode_json($content)->{$type} };
			$player->currentPlaylistUpdateTime( Time::HiRes::time() );
			if ($@) { # if error
				$log->error('error decoding [' . $content . '] or data [' . $type . '] not found');
				$log->error($@);
				return {};
			} else {
				### OK, we got this, here's your information
				$log->debug('information was obtained from the zone, storing in cache');
				$cache->set($hashkey, $resp, $cacheInfo{$type}{'duration'});
				$player->currentPlaylistUpdateTime( Time::HiRes::time() );
				return $resp;
			}
		} else {
			$log->debug("we will do this async, we have in our caching queue [" . scalar @caching . "] requests being processed...");
			### We can do this async! Let's check if we are already caching this.
			my $inProgress = 0;
			foreach (@caching) {
				if ($hashkey==$_) {
					$inProgress = 1;
					$log->debug("we already have this one in progress!");
				}
			}
			$log->debug("do we have it in cache? [" . $inProgress . "]");
			if ( $inProgress ) {
				### We have this in our list of stuff to get, be patient, eat a cookie and come back later...
				$log->info("I think we are already caching $hashkey, let's try later");
				### ... you only get an empty hash this time.
				return {};
			}
			### OK, we need to put this in our queue. Please come back later...
			$log->info('try filling the cache async!');
			fill_cache($ip, $hashkey, $type, $url);
			$log->debug('return empty hash for now');
			### ... for now, you can have this nice empty hash!
			return {}
		}
	}
}

sub fill_cache {
	my ($ip, $hashkey, $type, $url) = @_;
	$log->info('doing an async cache fill');
	push @caching, $hashkey;
	$log->debug(Dumper(@caching));
	callVzcAsync($ip, $type, $url,
		\&saveCache,
		\&fillCacheError,
		{	type => $type,
			hashkey => $hashkey,
			timestamp => Time::HiRes::time(),
		});
}

sub saveCache {
	my $http = shift;
	my $params   = $http->params;
	my $type = $params->{type};
	my $hashkey = $params->{hashkey};
	my $timestamp = $params->{timestamp};
	my $endTime = Time::HiRes::time();
	$log->info("we have info to put in the cache for $hashkey");
	$log->info("succesfully downloaded info in [" .  ($endTime - $timestamp) . "] secs");

	my $content = $http->content;
	eval { $resp = decode_json($content)->{$type} };
	if ($@) { # if error
		$log->error('error decoding [' . $content . '] or data [' . $type . '] not found');
		$log->error($@);
	} else {
		$log->info('information was obtained from the zone, storing in cache');
		$cache->set($hashkey, $resp, $cacheInfo{$type}{'duration'});
		removeFromList($hashkey);
	}
}

sub fillCacheError {
	my $http = shift;
	my $params   = $http->params;
	my $hashkey = $params->{hashkey};
	my $timestamp = $params->{timestamp};
	my $endTime = Time::HiRes::time();
	$log->error('*** error filling cache ***');
	$log->error("fill cache failed after [" .  ($endTime - $timestamp) . "] secs");	
	removeFromList($hashkey);
}

sub removeFromList {
	my $hashkey = shift;
	my $index = 0;
	$index++ until $caching[$index] eq $hashkey;
	splice(@caching, $index, 1);
}


1;

